`default_nettype none

module bmpparser(clk, reset, tx, rx, led);
    parameter sys_clk = 40000000;
    parameter rate    = 19200;

    input  wire clk;   // ̓NbN
    input  wire reset; // ZbgMD'1'̂ƂɉHZbg
    output wire tx;    // VAM̏o
    input  wire rx;    // VAM̓
    output wire [3:0] led;

    // ubNRAMCX^XG[V
    wire [17:0] mem_waddr, mem_raddr;
    wire [ 7:0] mem_wdata, mem_rdata;
    wire mem_we;
    ram16x8 ram0(.clock(clk), .wren(mem_we), .wraddress(mem_waddr[15:0]), .data(mem_wdata), 
                 .rdaddress(mem_raddr[15:0]), .q(mem_rdata));

    // RS232C̎MW[̃CX^X𐶐
    wire [7:0] rs232c_rdata;
    wire rs232c_rd;
    rs232c_rx#(.sys_clk(sys_clk), .rate(rate))
    u_rx(.clk(clk), .rst(reset), .din(rx), .rd(rs232c_rd), .dout(rs232c_rdata));

    // RS232C̑MW[̃CX^X𐶐
    reg [7:0] rs232c_wdata;
    reg rs232c_wr;
    wire rs232c_ready;
    rs232c_tx#(.sys_clk(sys_clk), .rate(rate))
    u_tx(.clk(clk), .rst(reset), .wr(rs232c_wr), .din(rs232c_wdata), .dout(tx), .ready(rs232c_ready));

    // O[XP[W[̃CX^X𐶐
    reg  grayscale_req;
    wire grayscale_we, grayscale_done;
    wire [19:0] grayscale_waddr, grayscale_raddr;
    wire [31:0] grayscale_wdata, grayscale_rdata;
    wire [31:0] grayscale_offset, grayscale_size;
    grayscale u(.clk(clk), .reset(reset), .req(grayscale_req), .size(grayscale_size), .offset(grayscale_offset),
                .we(grayscale_we), .waddr(grayscale_waddr), .wdata(grayscale_wdata),
                .raddr(grayscale_raddr), .rdata(mem_rdata), .done(grayscale_done));

    // ϐ̐錾
    reg [2:0] state, substate;
    reg [31:0] i;
    reg rd_d;

    reg [7:0] wdata;
    reg [31:0] waddr, raddr;
    reg we;

    reg [31:0] offset, width, height;
    reg [31:0] size;

    assign grayscale_offset = offset;
    assign grayscale_size = width * height;

    assign mem_waddr = (state == 0 || state == 1) ? waddr : grayscale_waddr;
    assign mem_wdata = (state == 0 || state == 1) ? wdata : grayscale_wdata;
    assign mem_we    = (state == 0 || state == 1) ? we    : grayscale_we;

    assign mem_raddr = (state == 5) ? raddr : grayscale_raddr;

    assign led = state == 0 ? 4'b0001 :
                 state == 1 ? 4'b0010 :
                 state == 2 ? 4'b0011 :
                 state == 3 ? 4'b0100 :
                 state == 4 ? 4'b0101 :
                 state == 5 ? 4'b0110 :
                 4'b1111;

    always @(posedge clk) begin
        if(reset == 1'b1) begin
            state <= 0;
            rd_d <= 1'b0;
            i <= 0;
            grayscale_req <= 1'b0;
            size <= 0;
        end else begin
            rd_d <= rs232c_rd;
            case(state)
                0: begin // f[^̎MJn܂őҋ@
                    if(rd_d == 1'b0 && rs232c_rd == 1'b1) begin // f[^̎MJn
                        state <= state + 1;
                        we    <= 1'b1;
                        wdata <= rs232c_rdata;
                        waddr <= 0;
                        i <= 1;
                    end else begin
                        we <= 1'b0;
                        i <= 0;
                    end
                    grayscale_req <= 1'b0;
                end
                1: begin // VAf[^M
                    if(rd_d == 1'b0 && rs232c_rd == 1'b1) begin // f[^M
                        i     <= i + 1;
                        waddr <= i;
                        wdata <= rs232c_rdata;
                        we <= 1'b1;
                        case(i)
                            // size
                             2: size[ 7: 0] <= rs232c_rdata;
                             3: size[15: 8] <= rs232c_rdata;
                             4: size[23:16] <= rs232c_rdata;
                             5: size[31:24] <= rs232c_rdata;
                            // offset
                            10: offset[ 7: 0] <= rs232c_rdata;
                            11: offset[15: 8] <= rs232c_rdata;
                            12: offset[23:16] <= rs232c_rdata;
                            13: offset[31:24] <= rs232c_rdata;
                            // width
                            18: width[ 7: 0] <= rs232c_rdata;
                            19: width[15: 8] <= rs232c_rdata;
                            20: width[23:16] <= rs232c_rdata;
                            21: width[31:24] <= rs232c_rdata;
                            // height
                            22: height[ 7: 0] <= rs232c_rdata;
                            23: height[15: 8] <= rs232c_rdata;
                            24: height[23:16] <= rs232c_rdata;
                            25: height[31:24] <= rs232c_rdata;
                        endcase // case (i)
                    end else begin
                        we <= 1'b0;
                    end
                    if((i >= size) && (i > 5)) begin
                        i     <= 0;
                        state <= state + 1;
                    end
                end
                2: begin // O[XP[W[N
                    grayscale_req <= 1'b1;
                    state <= state + 1;
                end
                3: begin
                    grayscale_req <= 1'b0;
                    if(!grayscale_done) begin // ̊tOfalseɂȂ玟̃Xe[gɂ
                        state <= state + 1;
                    end
                end
                4: begin
                    if(grayscale_done) begin // ̊tOtrueɂȂ玟̃Xe[gɂ
                        i <= 0;
                        state <= state + 1;
                        substate <= 0;
                    end
                end
                5: begin
                    if(i >= size) begin // SM
                        state <= 0;
                    end else begin
                        case(substate)
                            0: begin
                                raddr <= i; // Mf[^̃AhXw
                                substate <= substate + 1;
                            end
                            1: begin        // 1NbN҂
                                substate <= substate + 1;
                            end
                            2: begin
                                if(rs232c_ready == 1'b1) begin // RS232CMW[ɃNGXgs\
                                    rs232c_wdata <= mem_rdata;
                                    rs232c_wr <= 1'b1; // NGXg𔭍s
                                    substate <= substate + 1;
                                end
                            end
                            3: begin
                                if(rs232c_ready == 1'b0) begin // RS232CMW[ɃNGXg󗝂ꂽ
                                    rs232c_wr <= 1'b0;
                                    substate  <= 0;
                                    i <= i + 1;
                                end
                            end
                        endcase
                    end
                end
            endcase
        end
    end



endmodule // bmpparser

`default_nettype wire
